#include <stdio.h>
#include <stdlib.h>
#include <Windows.h>
#include <Psapi.h>

#define MAX_NUM_PROCESSES 1024

static inline void exitWithErrorNoMemory(PVOID assertVal)
{
    if (!assertVal) {
        printf("Not enough memory, exiting...\n");
        ExitProcess(ENOMEM);
    }
}

static DWORD retrieveProcesses(PDWORD *processesIdArray, DWORD arraySize)
{
    DWORD bytesNeeded = 0;
    BOOL ret;

    ret = EnumProcesses(*processesIdArray, arraySize, &bytesNeeded);
    if (ret == FALSE) {
        return 0;
    }
    return bytesNeeded;
}

static HANDLE openProcess(DWORD processId)
{
    return OpenProcess(PROCESS_ALL_ACCESS, FALSE, processId);
}

static void closeProcessList(PHANDLE processHandles, DWORD handleArraySize)
{
    for (DWORD i = 0; i < handleArraySize; ++i) {
        CloseHandle(processHandles[i]);
    }
}

static DWORD processesToHandles(PHANDLE handleArray)
{
    PDWORD processIdArray = (PDWORD) malloc(MAX_NUM_PROCESSES * sizeof(DWORD));
    DWORD processIdSize;
    HANDLE currentHandle;
    DWORD openHandles = 0;
    exitWithErrorNoMemory(processIdArray);
    processIdSize = retrieveProcesses(&processIdArray, MAX_NUM_PROCESSES);

    for (DWORD i = 0; i < processIdSize; ++i) {
        if (processIdArray[i] >= 0) {
            currentHandle = openProcess(processIdArray[i]);
            if (currentHandle > 0) {
                handleArray[openHandles] = currentHandle;
                openHandles++;
            }
        }
    }

    free(processIdArray);
    return openHandles;
}

static void getMemoryInfoForHandles(PHANDLE processArray, PPROCESS_MEMORY_COUNTERS counters, DWORD processArraySize)
{
    for (DWORD i = 0; i < processArraySize; ++i) {
        GetProcessMemoryInfo(processArray[i], &counters[i], sizeof(*counters));
    }
}

static PCHAR *getProcessNameForHandles(PHANDLE processArray, DWORD processArraySize)
{
    PCHAR *paths = (PCHAR *) malloc(processArraySize * sizeof(*paths));
    DWORD size = MAX_PATH;
    exitWithErrorNoMemory(paths);

    for (DWORD i = 0; i < processArraySize; ++i) {
        paths[i] = (PCHAR) malloc(MAX_PATH * sizeof(**paths));
        exitWithErrorNoMemory(paths[i]);
        size = MAX_PATH;
        QueryFullProcessImageNameA(processArray[i], PROCESS_NAME_NATIVE, paths[i], &size);
    }

    return paths;
}

static PERFORMANCE_INFORMATION getSystemInfo()
{
    PERFORMACE_INFORMATION perf;
    GetPerformanceInfo(&perf, sizeof(perf));
    return perf;
}

static int extractExecutable(PCHAR pathName)
{
    for (int i = strlen(pathName) - 1; i >= 0; --i) {
        if (pathName[i] == '\\') {
            return i + 1;
        }
    }
    return 0;
}

static void printSystemInfo(PERFORMANCE_INFORMATION perfInfo)
{
    printf("CommitTotal, CommitLimit, CommitPeak, PhysicalTotal, PhysicalAvailable, SystemCache, "
        "KernelTotal, KernelPaged, KernelNonpaged, PageSize, HandleCount, ProcessCount, ThreadCount\n");
    printf("%d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d\n\n",
        perfInfo.CommitTotal,
        perfInfo.CommitLimit,
        perfInfo.CommitPeak,
        perfInfo.PhysicalTotal,
        perfInfo.PhysicalAvailable,
        perfInfo.SystemCache,
        perfInfo.KernelTotal,
        perfInfo.KernelPaged,
        perfInfo.KernelNonpaged,
        perfInfo.PageSize,
        perfInfo.HandleCount,
        perfInfo.ProcessCount,
        perfInfo.ThreadCount
        );
}

static void printTableInfo(PPROCESS_MEMORY_COUNTERS memoryCounterData, PCHAR *processNames, DWORD counterSize)
{
    printf("ProcessName, PageFaultCount, PeakWorkingSetSize, WorkingSetSize, QuotaPeakPagedPoolUsage, QuotaPagedPoolUsage, "
            "QuotaPeakNonPagedPoolUsage, QuotaNonPagedPoolUsage, PagefileUsage, PeakPagefileUsage\n");

    for (DWORD i = 0; i < counterSize; ++i) {
        printf("%s, %d, %d, %d, %d, %d, %d, %d, %d, %d\n",
            processNames[i] + extractExecutable(processNames[i]),
            memoryCounterData[i].PageFaultCount,
            memoryCounterData[i].PeakWorkingSetSize / 1024,
            memoryCounterData[i].WorkingSetSize / 1024,
            memoryCounterData[i].QuotaPeakPagedPoolUsage / 1024,
            memoryCounterData[i].QuotaPagedPoolUsage / 1024,
            memoryCounterData[i].QuotaPeakNonPagedPoolUsage / 1024,
            memoryCounterData[i].QuotaNonPagedPoolUsage / 1024,
            memoryCounterData[i].PagefileUsage / 1024,
            memoryCounterData[i].PeakPagefileUsage / 1024
            );
    }
}

static void initEnvironment(PHANDLE *handleArray)
{
    *handleArray = (PHANDLE) malloc(MAX_NUM_PROCESSES * sizeof(**handleArray));
}

static void cleanupEnvironment(PHANDLE handleArray, PPROCESS_MEMORY_COUNTERS memoryCounterData, PCHAR *processNames, DWORD openProcesses)
{
    for (DWORD i = 0; i < openProcesses; ++i) {
        CloseHandle(handleArray[i]);
        free(processNames[i]);
    }
    free(handleArray);
    free(memoryCounterData);
    free(processNames);
}

int main()
{
    PHANDLE handleArray;
    DWORD openProcesses;
    PPROCESS_MEMORY_COUNTERS memoryCounterData;
    PERFORMANCE_INFORMATION perfInfo;
    PCHAR *processNames;

    initEnvironment(&handleArray);
    openProcesses = processesToHandles(handleArray);
    memoryCounterData = (PPROCESS_MEMORY_COUNTERS) calloc(openProcesses, sizeof(PROCESS_MEMORY_COUNTERS));
    exitWithErrorNoMemory(memoryCounterData);

    getMemoryInfoForHandles(handleArray, memoryCounterData, openProcesses);
    processNames = getProcessNameForHandles(handleArray, openProcesses);
    perfInfo = getSystemInfo();
    printSystemInfo(perfInfo);
    printTableInfo(memoryCounterData, processNames, openProcesses);
    cleanupEnvironment(handleArray, memoryCounterData, processNames, openProcesses);
    return 0;
}
